setClass("PointProcessModel",
         representation(
                        processData = "ProcessData",
                        modelMatrix = "Matrix",
                        delta = "numeric",
                        formula = "formula",
                        family = "Family",
                        coefficients = "numeric",
                        fixedCoefficients="list",
                        Omega = "matrix",
                        penalization = "logical",
                        var = "matrix",
                        call="call",
                        optimResult="list"
                        )
         )

setMethod("initialize","PointProcessModel",
          function(.Object,
                   processData,
                   formula,
                   family,
                   fit=TRUE,
                   coefficients=NULL,
                   fixedCoefficients=list(),
                   modelMatrix=TRUE,
                   fisherInformation=modelMatrix,
                   Omega=NULL,
                   call=NULL,...){
            .Object@processData <- processData
            .Object@formula <- formula
            .Object@family <- family
            if(!is.null(Omega)) {
              if(any(Omega != t(Omega))) {
                Omega <- (Omega + t(Omega))/2
                warning("Penalization matrix Omega is not symmetric and is replaced by  (Omega + t(Omega))/2.")
              }
              if(any(eigen(Omega,only.values=TRUE,symmetric=TRUE)$values < 0)) {
                stop("Penalization matrix Omega is not positive semi-definit.")
              }
              .Object@Omega <- Omega
              .Object@penalization <- TRUE
            } else if(is.null(Omega)) .Object@penalization <- FALSE
           
            .Object@delta <- as.numeric(unlist(tapply(getPosition(getContinuousProcess(processData)),
                                                      getId(getContinuousProcess(processData)),
                                                      function(x) c(diff(x),0)),use.names=FALSE))

            if(modelMatrix) {
              .Object <- computeModelMatrix(.Object)
              
## Deprecated code              
##              if(fisherPreComp) {
##                .Object@XtX <-  lapply(seq(1,dim(.Object@modelMatrix)[2]),
##                                       function(i) .Object@modelMatrix[,i] * .Object@modelMatrix)
##              }

              parDim <- dim(.Object@modelMatrix)[2]

              if(is.null(coefficients)){
                coefficients <- rep(0,parDim)
              } else {
                if(length(coefficients) != parDim) {
                  coefficients <- rep(0,parDim)
                  warning("Incorrect dimension of initial parameter vector. Initial parameters all set to 0")
                }
              }

              if(length(fixedCoefficients) != 0) coefficients[fixedCoefficients$which] <- fixedCoefficients$value
              .Object@fixedCoefficients <- fixedCoefficients
              .Object@coefficients <- coefficients
            
              if(fit){
                .Object <- glppmFit(.Object,coefficients,fisherInformation=fisherInformation,...)
##                This is also implemented in glppmFit. One may be concerned with
##                too much data copying by calling glppmFit at this point, but it seems
##                easier to maintain code if we have the actual optimization in one place only.
##                The code below may be outdated - compare to glppmFit.
##                mll <- function(par,...) computeMinusLogLikelihood(.Object,par,...)
##                dmll <- function(par,...) computeDMinusLogLikelihood(.Object,par,...)
                
##                 .Object@optimResult <- optim(coefficients,mll,gr=dmll,method="BFGS",...)
##                 .Object@coefficients <- .Object@optimResult$par
##                 names(.Object@coefficients) <- dimnames(.Object@modelMatrix)[[2]]
##                 if(fisherInformation) {
##                   vcov <- solve(computeDDMinusLogLikelihood(.Object))
##                   rownames(vcov) <- names(.Object@coefficients)
##                   colnames(vcov) <- names(.Object@coefficients)
##                   .Object@var <- (vcov + t(vcov))/2   ## To assure symmetry
##                }
              }
            }
            
            if(!is.null(call)) .Object@call <- call
            return(.Object)

          } 
        )

setMethod("subset","PointProcessModel",
          function(x,...){
            pointProcessModel <- new("PointProcessModel",
                                     processData=subset(x@processData,...),
                                     formula=object@formula,
                                     family=object@family)
            return(pointProcessModel)
          }
          )



setMethod("getModelMatrix","PointProcessModel",
          function(object,...){
            return(object@modelMatrix)
          }
          )


setMethod("coefficients","PointProcessModel",
          function(object,...){
            return(object@coefficients)
          }
          )

setReplaceMethod("coefficients","PointProcessModel",
          function(.Object,value){
            .Object@coefficients <- value
            return(.Object)
          }
          )

setMethod("vcov","PointProcessModel",
          function(object,...){
            return(object@var)
          }
          )

                   
setMethod("formula","PointProcessModel",
          function(x,...){
            return(x@formula)
          }
          )

setReplaceMethod("formula","PointProcessModel",
          function(.Object,value){
                        
            if(is.formula(value)) {
              .Object@formula <- value
            } else stop("formula needs to be a 'formula'")

            return(.Object)
          }
          )


### calls to model.matrix should be replaced by calls to sparse.model.matrix
### from R.2.10. 


setMethod("computeModelMatrix","PointProcessModel",
          function(object,evaluationPositions=NULL,...){

            designList <- list()
            design <- list()

            if(is.null(evaluationPositions)) {
              evalPositions <- tapply(getPosition(getContinuousProcess(object@processData)),
                                      getId(getContinuousProcess(object@processData)),list)
            } else {
              evalPositions <- evaluationPositions
            }
            
            markedPointProcess <- getMarkedPointProcess(object@processData)
            positions <- getPosition(markedPointProcess)

            id <- factor(getId(markedPointProcess))
            idLevels <- levels(id)

            marks <- getMarkType(markedPointProcess)
            markLevels <- levels(marks)

            mt <- terms(object@formula) 
            termLabels <- attr(mt,"term.labels")
            notMarkTerms <- character()

            ### Model matrix computations for the terms involving 
            ### the marks:
            
            for(term in termLabels){
              form <- as.formula(paste("~",term,"-1"))
              variables <- all.vars(form)
              mark <- markLevels[markLevels %in% variables]
              if(length(mark) > 1) {
                stop(paste("Interaction of two mark types in '",
                           term,"' is currently not implemented.",sep=""))
              } else if(length(mark) == 1) {
                posi <- positions[marks==mark]
                idi <- id[marks==mark]
                idList <- factor()
                last <- as.list(rep(0,length(idLevels)))
                names(last) <- idLevels
                integrationMatrix <- list()

                if(length(variables) == 1) { 
                  for(j in seq(along=posi)){
                    afterPosi <- evalPositions[[idi[j]]] > posi[j]
                    val <- list(evalPositions[[idi[j]]][afterPosi] - posi[j])
                    names(val) <- mark
                    tmp <- model.matrix(form,val)
                    if(!(idi[j] %in% idLevels[idList])) {
                      designList[[paste(term,idi[j],sep="")]] <-  Matrix(0,nrow=length(evalPositions[[idi[j]]]),
                                                                         ncol=length(which(idi == idi[j]))*dim(tmp)[2],
                                                                         sparse=TRUE)
                      integrationMatrix[[as.character(idi[j])]] <-  Matrix(rep(diag(1,dim(tmp)[2]),length(which(idi==idi[j]))),
                                                                           ncol=dim(tmp)[2],byrow=TRUE,sparse=TRUE,
                                                                           dimnames=list(rownames=NULL,colnames=colnames(tmp)))
                      idList <- c(idList,idi[j])
                    }
                    designList[[paste(term,idi[j],sep="")]][afterPosi,
                                                         seq(last[[idi[j]]]+1,
                                                             last[[idi[j]]]+dim(tmp)[2])] <- tmp
                    last[[idi[j]]] <- last[[idi[j]]]+dim(tmp)[2]                   
                    }
                } else if(length(variables > 1)) {
                  if(!(all(variables[variables != mark] %in% names(getMarkValue(markedPointProcess))))){
                    stop(paste("Use of one (or more) variables in '",term,"' that does not exist",sep=""))
                  } else {
                    markValues <- subset(getMarkValue(markedPointProcess),marks==mark)
                    for(j in seq(along=posi)){
                    afterPosi <- evalPositions[[idi[j]]] > posi[j]
                    val <- list(evalPositions[[idi[j]]][afterPosi] - posi[j])
                    val[2:(1+dim(markValues)[2])] <- as.list(markValues[j,])
                    names(val) <-c(mark,names(markValues))
                    tmp <- model.matrix(form,val)
                    tmp <- Matrix(tmp,dimnames=dimnames(tmp))
                    if(!(idi[j] %in% idLevels[idList])) {
                      designList[[paste(term,idi[j],sep="")]] <-  Matrix(0,nrow=length(evalPositions[[idi[j]]]),
                                                                      ncol=length(which(idi == idi[j]))*dim(tmp)[2],
                                                                      sparse=TRUE)
                      integrationMatrix[[as.character(idi[j])]] <-  Matrix(rep(diag(1,dim(tmp)[2]),length(which(idi==idi[j]))),
                                                                           ncol=dim(tmp)[2],byrow=TRUE,sparse=TRUE,
                                                                           dimnames=list(rownames=NULL,colnames=colnames(tmp)))
                      idList <- c(idList,idi[j])
                    }
                    designList[[paste(term,idi[j],sep="")]][afterPosi,
                                                         seq(last[[idi[j]]]+1,
                                                             last[[idi[j]]]+dim(tmp)[2])] <- tmp
                    last[[idi[j]]] <- last[[idi[j]]]+dim(tmp)[2]
                    }
                  }
                }
              
                if(idLevels[1] %in% idLevels[idList]) {
                  design[[term]] <-  designList[[paste(term,idLevels[1],sep="")]] %*% integrationMatrix[[idLevels[1]]]
              } else {
                design[[term]] <- Matrix(0,nrow=length(evalPositions[[idLevels[1]]]),ncol=dim(tmp)[2],sparse=TRUE)
              }
                for(k in seq(along=idLevels)[-1]){
                  if(idLevels[k] %in% idLevels[idList]) {               
                    design[[term]] <- rBind(design[[term]],
                                         designList[[paste(term,idLevels[k],sep="")]] %*% integrationMatrix[[idLevels[k]]])   
                  } else {
                    design[[term]] <- rBind(design[[term]],
                                         Matrix(0,nrow=length(evalPositions[[idLevels[k]]]),ncol=dim(tmp)[2],sparse=TRUE))
                  }
                }
              } else if(length(mark) == 0) notMarkTerms <- c(notMarkTerms,term)
            }

            ### Model matrix computations for the id and continuous time process
            ### components

            if(attr(mt,"intercept") == 1){
              notMarkTerms <- paste(c(notMarkTerms,"1"),collapse="+")
            } else if(length(notMarkTerms) >0) {
              notMarkTerms <-  paste(paste(notMarkTerms,collapse="+"),"-1")
            }

            if(length(notMarkTerms) > 0){
              form <-  as.formula(paste("~",notMarkTerms))
              variables <- all.vars(form)
              if(all(variables %in% c("id",colnames(getValue(getContinuousProcess(object@processData)))))) {
                values <- data.frame(id=getId(getContinuousProcess(object@processData)))
                notIdVariables <- variables[variables != "id"]
                if(length(notIdVariables) > 0) {
                  tmp <- as.matrix(getValue(getContinuousProcess(object@processData))[,notIdVariables,drop=FALSE])
                  rownames(tmp) <- rownames(values)
                  values <- cbind(values,tmp)
                }
                tmp <- model.matrix(form,values)
                X0 <- Matrix(tmp,dimnames=dimnames(tmp))
###                X0 <- sparse.model.matrix(form,values)
              } else {
                stop(paste("Use of non existing variable(s) in:", form))
              }
            }
          
          

            Dim2 <- function(x) {
              if(!is.null(dim(x))) result <- dim(x)[2]
              if(is.null(dim(x))) result <- 1
              return(result)
            }
            
            designDim <- c(sum(sapply(evalPositions,length)),
                           if(identical(design,list())) 0 else {
                             sum(sapply(design,Dim2))})

            if(length(notMarkTerms) > 0) designDim[2] <- designDim[2] + dim(X0)[2]
            X <- Matrix(0,nrow=designDim[1],ncol=designDim[2],sparse=TRUE)
            X[,seq(1,dim(X0)[2])] <- X0
            last <- dim(X0)[2]
            for(i in seq(along=design)) {
                X[,seq(last+1,last+Dim2(design[[i]]))] <- design[[i]]
                last <-  last+Dim2(design[[i]])
            }

            dimnames(X) <- list(rownames=NULL,colnames=c(colnames(X0),unlist(lapply(design,colnames))))
            object@modelMatrix <- X
            
            return(object)
          }
          )
            
setMethod("computeLinearPredictor","PointProcessModel",
          function(object,coefficients=NULL,...){
            if(is.null(coefficients)) {
              coefficients <- coefficients(object)
            } 
                                     
            eta =  as.numeric(object@modelMatrix %*% coefficients)
            return(eta)
          }
          )

setMethod("computeMinusLogLikelihood","PointProcessModel",
          function(object,coefficients,...){
            eta <- computeLinearPredictor(object,coefficients)
             if(attr(terms(object@formula),"response") != 0) {
              response <- all.vars(object@formula,unique=FALSE)[attr(terms(object@formula),"response")]
            } else stop("no response variable specified")
            

            if(object@family@link == "log"){
              mll <-  sum(exp(eta)*object@delta) -
                sum(eta[getMarkTypePosition(object@processData,response)]) 
            } else {
              mll <-  sum(object@family@phi(eta)*object@delta) -
                sum(log(object@family@phi(eta[getMarkTypePosition(object@processData,response)]))) 
            }
            
            return(mll)
            
          }
          )

setMethod("update","PointProcessModel",
          function(object,...){
            .local <- function(object,formula,fisherInformation=TRUE,warmStart=TRUE,fixedPar=list(),...){

              updatedFormula <- update(object@formula,formula)
              object@formula <- updatedFormula
              initPar <- NULL

              updatedTerms <- attr(terms(updatedFormula),"term.labels")
              if(attr(terms(updatedFormula),"intercept")==1) updatedTerms <- c(updatedTerms,"Intercept")

              updatedModelMatrixColumns <- sapply(updatedTerms,grep,colnames(object@modelMatrix),fixed=TRUE,simplify=FALSE)
                            
              if(any(sapply(updatedModelMatrixColumns,function(x) length(x) == 0))) {
                object <- computeModelMatrix(object)
                object@coefficients <- rep(0,dim(object@modelMatrix)[2])
              } else {
                col <- sort(unlist(updatedModelMatrixColumns))
                if(any(attr(terms(object@formula),"order") >= 2)) {
                  warning("Original model specification includes interaction terms. The updated model may not be as desired.")
                }
                object@modelMatrix <- object@modelMatrix[,col]
##                object@XtX <- lapply(seq(along=object@XtX),function(i) object@XtX[[i]][,col]) #Deprecated
                if(warmStart) object@coefficients <- initPar <- coefficients(object)[col]
              }

              if(length(fixedPar) != 0) object@coefficients[fixedPar$which] <- fixedPar$value
              object@fixedCoefficients <- fixedPar
                          
              return(glppmFit(object,initPar=initPar,fisherInformation=fisherInformation,...))
              
            }
            object@call <- match.call()
            .local(object,...)
          }
          )


setMethod("getLinearFilter",c(x="numeric",object="PointProcessModel"),
          function(x,object,...){

            mpp <- getMarkedPointProcess(object@processData)
            evaluations <- as.data.frame(x)
            termLabels <- attr(terms(object@formula),"term.labels")
            i <- 1
            linearFilter <- list()
            linearFilterSE <- list()
            
            for(term in termLabels){
              form <- as.formula(paste("~",term,"-1"))
              if(all( all.vars(form) %in% c(levels(getMarkType(mpp)),colnames(getMarkValue)))){
                names(evaluations) <- all.vars(form)
                design <- model.matrix(form,evaluations)
                linearFilter[[all.vars(form)]] <- design %*% coefficients(object)[dimnames(design)[[2]]]
                linearFilterSE[[all.vars(form)]] <- sqrt(diag(design %*% vcov(object)[dimnames(design)[[2]],dimnames(design)[[2]]] %*% t(design)))
              }
            }

            list(linearFilter=cbind(x, as.data.frame(linearFilter)),se=linearFilterSE)
          }
          )

setMethod("plot",signature(x="numeric",y="PointProcessModel"),
          function(x,y,trans=NULL,alpha=0.05,...){
            linearFilter <- getLinearFilter(x,y)
            moltenFilter <- melt(linearFilter$linearFilter,id.vars="x")
            q <- qnorm(1-alpha/2)
            plotData <- cbind(moltenFilter,data.frame(cf.lower = moltenFilter$value-q*unlist(linearFilter$se),
                                                      cf.upper = moltenFilter$value+q*unlist(linearFilter$se)))
                              
            if(!is.null(trans)) plotData[,c("value","cf.lower","cf.upper")] <- do.call(trans,list(plotData[,c("value","cf.lower","cf.upper")]))

            linearFilterPlot <- ggplot(data=plotData,aes(x=x,y=value)) +
              facet_grid(.~variable) +
                geom_ribbon(aes(min=cf.lower,max=cf.upper),fill=alpha("blue",0.2)) +
                  scale_x_continuous("position") +
                    scale_y_continuous("") + geom_line()

            return(linearFilterPlot)
          }
          )
          
setMethod("computeDMinusLogLikelihood","PointProcessModel",
          function(object,coefficients,...){
            eta <- computeLinearPredictor(object,coefficients)
             if(attr(terms(object@formula),"response") != 0) {
              response <- all.vars(object@formula,unique=FALSE)[attr(terms(object@formula),"response")]
            } else stop("no response variable specified")

            if(object@family@link == "log"){
              dmll <-  colSums((exp(eta)*object@delta)*object@modelMatrix) -
                colSums(object@modelMatrix[getMarkTypePosition(object@processData,response),]) 
            } else {
              etaP <- eta[getMarkTypePosition(object@processData,response)]
              mmP <- object@modelMatrix[getMarkTypePosition(object@processData,response),]
              dmll <-  colSums((object@family@Dphi(eta)*object@delta)*object@modelMatrix) -
                colSums(object@family@Dphi(etaP)/object@family@phi(etaP)*mmP) 
            }
            
            return(dmll)
            
          }
          )

setMethod("computeDDMinusLogLikelihood","PointProcessModel",
          function(object,coefficients,...){
            eta <- computeLinearPredictor(object,coefficients)
             if(attr(terms(object@formula),"response") != 0) {
              response <- all.vars(object@formula,unique=FALSE)[attr(terms(object@formula),"response")]
            } else stop("no response variable specified")

##          Deprecated code
##             if(length(object@XtX)==0) {
##               if(object@family@link == "log"){
##                 d <- dim(object@modelMatrix)[2]
##                 weight <- exp(eta)*object@delta
##                 ddmll <- matrix(ncol=d,nrow=d)
##                 for(i in 1:d){
##                   ddmll[,i] <- colSums(weight * object@modelMatrix[,i] * object@modelMatrix)
##                 }
##               }
##             } else {      
##              
##              if(object@family@link == "log"){
##                ddmll <-  sapply(object@XtX,function(x) colSums((exp(eta)*object@delta)*x))
##              }
##            }

             if(object@family@link == "log"){
                ddmll <-  as(crossprod(object@modelMatrix,exp(eta)*object@delta*object@modelMatrix),"matrix")
              } else if(object@family@link == "identity"){
                etaP <- eta[getMarkTypePosition(object@processData,response)]
                mmP <- object@modelMatrix[getMarkTypePosition(object@processData,response),]
                ddmll <-  as(crossprod(mmP,1/object@family@phi(etaP)^2*mmP),"matrix")
                } else {
                etaP <- eta[getMarkTypePosition(object@processData,response)]
                mmP <- object@modelMatrix[getMarkTypePosition(object@processData,response),]
                ddmll <-  as(crossprod(object@modelMatrix,object@family@D2phi(eta)*object@delta*object@modelMatrix),"matrix") -
                  as(crossprod(mmP,(object@family@D2phi(etaP)*object@family@phi(etaP) - object@family@Dphi(etaP)^2)/object@family@phi(etaP)^2*mmP),"matrix")
               }
            
            return(ddmll)
            
          }
          )

setMethod("print","PointProcessModel",
          function(x,digits= max(3, getOption("digits") - 3), ...){
            cat("\nCall:\n", deparse(x@call), "\n\n", sep = "")
            if (length(coefficients(x))) {
              cat("Coefficients:\n")
              print.default(format(coefficients(x), digits = digits), print.gap = 2, 
                            quote = FALSE)
            }
            else cat("No coefficients\n")
            cat("\n")
            invisible(x)
          }
          )

setMethod("show","PointProcessModel",
          function(object) print(x=object)
          )


setMethod("summary","PointProcessModel",
          function(object,...) {
            result <- list()

            result$df <- length(object@coefficients) - length(object@fixedCoefficients$which)
            result$call <- object@call
            result$mll <- object@optimResult$value
            result$iter <- object@optimResult$counts
            result$convergence <- object@optimResult$convergence
            result$aic <- 2*(result$mll +  result$df)

            # Some way of summarizing 'residuals'

            se <- sqrt(diag(object@var))
            z <- object@coefficients/se
            q <- 2*(1-pnorm(abs(z)))
            z[se==0] <- NA
            q[se==0] <- NA
            se[se==0] <- NA
            
            result$coefficients <- matrix(c(object@coefficients,se,z,q),ncol=4)
            zapNames <- sapply(strsplit(names(object@coefficients),""),
                               function(x) {
                                 end <- x[seq(max(length(x)-2,1),length(x))]
                                 if(length(x) > 15) end <- c("...",x[seq(max(length(x)-2,1),length(x))])
                                 paste(c(x[1:min(max(length(x)-3,1),15)],end),sep="",collapse="")
                               })
                             
            
            dimnames(result$coefficients) <- list(zapNames,c("Estimate","Std. Error", "z value", "Pr(> |z|)"))

            result$fixedCoefficients <- object@fixedCoefficients
            result$penalization <- object@penalization
            class(result) <- c("summary.glppm")
            return(result)

          }
          )

          
